from lxml import etree
import re
import csv
from collections import deque

from sec_parsers.style_detection import detect_style_from_string, detect_style_from_element, detect_table,detect_link, detect_image,detect_table_of_contents, get_all_text, is_paragraph,\
detect_hidden_element
from sec_parsers.xml_helper import get_text, set_background_color, remove_background_color, open_tree, element_has_text,get_text_between_elements,get_elements_between_elements
from sec_parsers.visualization_helper import headers_colors_dict, headers_colors_list
from sec_parsers.hierarchy import get_hierarchy, get_preceding_elements, find_last_index
from sec_parsers.cleaning import clean_title
#TODO add better attributes, and a bunch of other stuff

# dictionary parsing_type = ignore is e.g. images that we currently ignore, text is added to previous node

# html attributes, parsing_string, parsing_type
def recursive_parse(element):
    """ Recursively parse elements to detect potential headers """
    
    # if element is invisible, skip
    if detect_hidden_element(element):
        return

    if detect_table(element):
        if detect_table_of_contents(element) == "toc":
            element.attrib['parsing_string'] = 'table of contents;'
        else:
            element.attrib['parsing_string'] = 'table;'
        return
    elif detect_link(element):
        element.attrib['parsing_string'] = 'link;'
        return
    elif detect_image(element):
        element.attrib['parsing_string'] = 'image;'
        return

    text = get_text(element).strip()
    if text == '':
        for child in element.iterchildren():
            recursive_parse(child)
    else:
        string_style = detect_style_from_string(text)
        element_style = detect_style_from_element(element)
        parsing_string = string_style + element_style

        # check if text element, e.g. not header.  WIP
        if parsing_string == '':
            return

        # determines whether to add parsing string to element or parent
        if ((get_all_text(element) == '') and (element.tail is not None)):
            parent = element.getparent()
            parent.attrib['parsing_string'] = parsing_string + 'parent;'
        # check if item has part of it in bold, rest as tail. e.g. 1606 Corp. 10K 2023 item 15
        elif (('item;' in parsing_string) and (element.tail is not None)):
            parent = element.getparent()
            parent.attrib['parsing_string'] = parsing_string #+ 'parent;'
        else:
            element.attrib['parsing_string'] = parsing_string

    return 

# WIP
from collections import deque

def relative_parsing(html):
    """Looks at parsed html from recursive parsing and uses relative position to improve parsing."""
    parsed_elements = deque(html.xpath('//*[@parsing_string]'))
    processed_elements = set()  # To keep track of processed elements

    while parsed_elements:
        parsed_element = parsed_elements.popleft()
        
        # Skip if this element has already been processed
        if parsed_element in processed_elements:
            continue
        
        parsing_string = parsed_element.get('parsing_string')

        # Process children
        children = parsed_element.xpath("child::*")
        if children:
            children_with_parsing_string = parsed_element.xpath('child::*[@parsing_string]')
            if children_with_parsing_string:
                parsed_element.attrib.pop('parsing_string')
            else:
                text = get_all_text(parsed_element)
                parsed_element.attrib['parsing_string'] = parsing_string + detect_style_from_string(text)

        # Process bullet points
        elif 'bullet point;' in parsing_string:
            if parsed_elements:
                next_element = parsed_elements[0]
                text_between = get_text_between_elements(html, parsed_element, next_element).strip()
                if text_between == '':
                    next_element.attrib.pop('parsing_string')
                    processed_elements.add(next_element)

        # Process headers and other elements
        elif parsing_string not in ['table of contents;', 'table;', 'link;', 'image;', 'signatures;']:
            if parsed_elements:
                next_element = parsed_elements[0]
                elements_between = get_elements_between_elements(html, parsed_element, next_element)
                flag = False
                if not elements_between:
                    flag = True
                # elif len(elements_between) == 1: # wip
                #     if elements_between[0].tag in ['div','span']:
                #         flag = True

                if flag: #  WIP struggles with some cases where whitespace between two elements. figure out how to fix w/o breaking other stuff
                    parent = parsed_element.getparent()
                    parent.attrib['parsing_string'] = parsing_string + "parent;"
                    next_element.attrib.pop('parsing_string')
                    parsed_element.attrib.pop('parsing_string')
                    processed_elements.add(next_element)
                    processed_elements.add(parsed_element)
                    parsed_elements.appendleft(parent)  # Add parent to be processed next
                else:
                    previous_element = parsed_element.getprevious()
                    if previous_element is not None: 
                        text = get_text(previous_element).strip() # WIP: may introduce some issues
                        if previous_element.get('parsing_string') is None and text != '':
                            parsed_element.attrib.pop('parsing_string', None)

        processed_elements.add(parsed_element)

    return html

def cleanup_parsing(html):
    # reads html and sets attributes
    parsed_elements = html.xpath('//*[@parsing_string]')
    for parsed_element in parsed_elements:
        parsing_string = parsed_element.get('parsing_string')
        parsing_type = parsing_string
        if any([item in parsing_string for item in ['table of contents;','link;','image;','page number;']]):
            parsing_type = 'ignore;'
        elif any([item in parsing_string for item in ['table']]):
            parsing_type = 'table;'
        elif 'part;' in parsing_string:
            parsing_type = 'part;'
        elif 'item;' in parsing_string:
            parsing_type = 'item;'
        elif 'signatures;' in parsing_string:
            parsing_type = 'signatures;'
        elif 'bullet point;' in parsing_string:
            parsing_type = None
        # elif any([item in parsing_string for item in ['font-style:italic;','em','i']]):
        #     text = get_text(parsed_element)
        #     if is_paragraph(text):
        #         parsing_type = None
            
        if parsing_type is not None:
            # remove parent
            # WIP
            parsing_type = parsing_type.replace('parent;','')
            parsed_element.attrib['parsing_type'] = parsing_type

    return html
        
def parse_10k(html):
    # recursive parse for style detection
    recursive_parse(html)

    # pruning based on relative locations
    relative_parsing(html)

    # standardizing certain elements such as part, item, signatures
    cleanup_parsing(html)

    return html

# 10q and 10k are the same for now
parse_10q = parse_10k

# TODO: make visualization easier to read at an instance by differentiating colors etc
def visualize(root):
    # remove style from all descendants so that background color can be set
    for descendant in root.iterdescendants():
        remove_background_color(descendant)

    # find all elements with parsing attribute
    elements = root.xpath('//*[@parsing_type]')
    # get all unique parsing values
    parsing_values = list(set([element.attrib['parsing_type'] for element in elements]))
    # create a color dict
    color_dict =dict(zip(parsing_values, headers_colors_list[:len(parsing_values)])) 
    # replace color dict values with values from headers_colors_dict
    for key in headers_colors_dict.keys():
        color_dict[key] = headers_colors_dict[key]
    for element in elements:
        # get attribute parsing
        parsing = element.attrib['parsing_type']
        if parsing == '':
            pass
        else:
            color = color_dict[parsing]
            set_background_color(element, color)

    open_tree(root)

# Heavily WIP
def construct_xml_tree(parsed_html):
    root = etree.Element('root')
    
    # find all parsing elements
    elements = parsed_html.xpath('//*[@parsing_type]')

    # find the first part parsing
    # WIP will be changed when introductory section parsing is added
    parts_elements = parsed_html.xpath("//*[@parsing_type='part;']")
    first_part_element = [element for element in parts_elements if re.match(r'^part\s+(i|1)$',get_all_text(element).strip(),re.IGNORECASE)][0]
    # find signature
    signatures = [element for element in elements if 'signatures;' in element.attrib['parsing_type']][0]

    # subset elements between first part and signature
    elements = elements[elements.index(first_part_element):elements.index(signatures)]

    # add the signatures to the end of the elements. we are not processing the signature right now, just adding it to the end as an anchor
    elements.append(signatures)

    element_parsing_types = [element.attrib['parsing_type'] for element in elements]

    # restrict certain headers
    restricted_headers = ['ignore;']
    # WIP
    element_parsing_types = [element_parsing_type for element_parsing_type in element_parsing_types if element_parsing_type not in restricted_headers]

    # get which headers are above which headers
    hierearchy = get_hierarchy(element_parsing_types)

    # start parsing
    node_list = []
    count = 0
    while count < len(elements)-1:
        element = elements[count]
        next_element = elements[count+1]

        # WIP
        # check if element is a text, if so, add to text of previous node
        element_parsing_type = element.attrib['parsing_type']
        if element_parsing_type == 'table;':
            node_list[-1].text += get_text_between_elements(parsed_html,element, next_element)
            count += 1
            continue
        # skip element
        elif element_parsing_type == 'ignore;':
            count += 1
            continue

        # construct node
        title = get_all_text(element)
        title = clean_title(title)
        text = get_text_between_elements(parsed_html,element, next_element)

        if element_parsing_type == 'part;':
            node_class = 'part'
        elif element_parsing_type == 'item;':
            node_class = 'item'
        elif element_parsing_type == 'signatures;':
            node_class = 'signatures'
        else:
            node_class = 'company_defined_section'

        node = etree.Element(node_class, title = title, parsing_type = element_parsing_type)
        node.text = text

        # should return a list of headers that are above the current header
        rulers = get_preceding_elements(hierearchy, element_parsing_type)

        # get the parsing strings of the nodes in the node_list
        node_parsing_types= [node.attrib['parsing_type'] for node in node_list]

        # find the last element in the node_parsing_strings which is in rulers
        index = find_last_index(node_parsing_types, rulers)
        if element_parsing_type == 'part;':
            root.append(node)
            node_list = [node]
        elif len(node_list) == 0:
            # add node to root 
            root.append(node)
            # add node to node_list
            node_list = [node]
        elif index == -1:
            # add node to last node in node_list
            node_list[-1].append(node)
            # add node to node_list
            node_list.append(node)
        else:
            # subset node_list to index
            node_list = node_list[:index+1]
            # add node to last node in node_list
            node_list[-1].append(node)
            # add node to node_list
            node_list.append(node)

        count += 1

    return root

def detect_filing_type(html):
    """TODO"""
    return "10K"

# Think about inheritance
class Parser:
    def __init__(self, html):
        self._setup_html(html)
        self.hierarchy = None # need to implement
        self.xml = None
        self.filing_type = None

    def _setup_html(self,html):
        parser = etree.HTMLParser(encoding='utf-8',remove_comments=True)
        parsed_html = etree.fromstring(html, parser)
        self.html = parsed_html

    # make util
    def _detect_filing_type(self):
        filing_type = detect_filing_type(self.html)
        self.filing_type = filing_type

    def _parse_10k(self):
        self.html = parse_10k(self.html)

    def _parse_10q(self):
        self.html = parse_10q(self.html)

    def set_filing_type(self, filing_type):
        self.filing_type = filing_type

    def parse(self):
        if self.filing_type is None:
            self._detect_filing_type()

        if self.filing_type == '10K':
            self._parse_10k()
        elif self.filing_type == '10Q':
            self._parse_10q()
        else:
            raise ValueError('Filing type not detected')
        
        self._to_xml()

    def visualize(self):
        visualize(self.html)

    def _to_xml(self):
        self.xml = construct_xml_tree(self.html)

    # functions to interact with xml

    # Find #
    def find_nodes_by_title(self,title):
        title = title.strip().lower()
        if self.xml is None:
            self.to_xml()

        # select all nodes with title attribute
        titles = self.xml.xpath(f"//*[@title]")
        # find all nodes with title
        nodes = [node for node in titles if title in node.attrib['title'].lower()]

        return nodes

    def find_nodes_by_text(self, text):
        """Find a node by text."""

        if self.xml is None:
            self.to_xml()

        return self.xml.xpath(f"//*[contains(text(), '{text}')]")
    
    def fuzzy_find_nodes_by_title(self,title):
        # TODO: implement
        pass
    # Interact with Node #

    # Note, needs refactor, also needs better spacing fix with text.
    def get_node_text(self,node):
        """Gets all text from a node, including title string."""
        text = ''
        text += node.attrib.get('title','') + '\n'

        node_text = node.text
        if node_text is not None:
            text += node.text + '\n'
            
        for child in node:
            text += self.get_node_text(child)
        
        return text
    
    # Interact with tree #

    # TODO: user friendly - e.g. if parsed not called call parse
    def get_tree(self,node=None, level=0):
        if node is None:
            node = self.xml
        tree_string = node.tag
        for child in node:
            tree_string += '\n' + '|-' * level + self.get_tree(child, level + 1)
        return tree_string
    
    def get_title_tree(self,node=None,level=0,attribute='title'):
        if node is None:
            node = self.xml
            
        tree_atrib = node.attrib.get(attribute,'')
        for child in node:
            tree_atrib += '\n' + '|-' * level + self.get_title_tree(child, level + 1,attribute)

        return tree_atrib
    
    # Save to file #
    def save_xml(self, filename):
        if self.xml is None:
            self.to_xml()

        with open(filename, 'wb') as f:
            f.write(etree.tostring(self.xml))

            
    def save_csv(self, filename):
        def get_rows(node,path):
            path = path + "/" + node.attrib.get('title')

            row_list = []

            row = {}
            row['path'] = path
            row['text'] = self.get_node_text(node)
            row_list.append(row)
            for child in node:
                row_list.extend(get_rows(child,path))

            return row_list
        
        row_list = []
        for child in self.xml.getchildren():
            row_list.extend(get_rows(child,''))
        
        keys = row_list[0].keys()

        with open(filename, 'w', newline='', encoding='utf-8') as output_file:
            dict_writer = csv.DictWriter(output_file, keys)
            dict_writer.writeheader()
            dict_writer.writerows(row_list)

    # TODO: implement
    def save_dta(self, filename):
        pass

        



import re

# Module to clean the text

# TODO

# NEW
# e.g. ITEM 1. BUSINESS --> item1, overview --> overview, Our Products and Services --> our_products_and_services

def clean_title(text): #  adjust to format items and parts into standardized format, e.g. no periods
    # add detection for item / other sections
    text = text.strip()
    text = re.sub("[^\S \t\n\r\f\v]+",' ',text)
    return text

# OLD
# WIP
def clean_text(text):
    text = text.strip()
    return text

# WIP
def clean_tag_name(text):
    text = text.lower()
    text = re.sub('\s+', '', text)
    return text


import requests
import random as rd

# Module to download the SEC filings
#TODO I'm happy with this file so far.

def random_num():
    return rd.randint(1, 1000000)


headers = {
    'User-Agent': f'Sample Company Name {str(random_num)} AdminContact@<sample company domain>.com'
}

def set_headers(name, email):
    """Set the headers for the requests."""
    global headers
    headers['User-Agent'] = f'{name} {email}'

def download_sec_filing(url):
    """Download the SEC filing from the given URL."""
    global headers

    # nudge to set your own headers
    if 'Sample Company Name' in headers['User-Agent']:
        print("Nudge: Please set your own headers using set_headers(name, email)")
    
    sec_response = requests.get(url, headers=headers)
    
    if sec_response.status_code != 200:
        print(f"Error {sec_response.status_code}: failed to download {url}")
        return None
    
    # important to encode the text as utf-8 as the lxml parser expects it
    html = sec_response.text.encode("utf-8")
    return html




# Important colors (strong contrast, emphasis)
headers_colors_dict = {'part;': '#B8860B',
                       'item;': '#BDB76B',
                       'bullet point;': '#FAFAD2', # LightGoldenRodYellow
                       'table;': '#FAFAD2', # LightGoldenRodYellow
                       'ignore;': '#E0FFFF', # Light Cyan
                          'signature;': '#E0FFFF', # Light Cyan
                       }

# Less important colors (muted tones, background elements)
headers_colors_list = [
    "#f5f5f5",  # Very Light Gray
    "#e0e0e0",  # Light Gray
    "#d6d6d6",  # Light Gray
    "#cccccc",  # Light Gray
    "#d9d9d9",  # Very Light Gray
    "#f0f0f0",  # Even Lighter Gray
    "#e6e6e6",  # Very Light Gray
    "#f2f2f2",  # Even Lighter Gray

    # Muted colors with lower saturation
    "#d3d3d3",  # Muted Gray
    "#d0cccc",  # Muted Blue
    "#e0e0e6",  # Muted Lavender
    "#e6e0e0",  # Muted Pink
    "#d9e0e0",  # Muted Teal
    "#e6dfd6",  # Muted Green
    "#e0d6d6",  # Muted Peach
    "#dcdcdc",  # Muted Silver

    # Even more muted colors
    "#e5e5e5",  # Very Light Gray (more muted)
    "#cfcfcf",  # Light Gray (more muted)
    "#bdbdbd",  # Light Gray (more muted)
    "#bebebe",  # Light Gray (more muted)
    "#cecece",  # Very Light Gray (more muted)
    "#ebeb eb",  # Even Lighter Gray (more muted)
    "#cfcfcf",  # Very Light Gray (more muted)
    "#d7d7d7",  # Even Lighter Gray (more muted)

    # Muted colors with even lower saturation (carefully consider contrast)
    "#d9d9d9",  # Muted Gray (even lower saturation)
    "#c2c2c2",  # Muted Blue (even lower saturation)
    "#d6d6d9",  # Muted Lavender (even lower saturation)
    "#d9d3d3",  # Muted Pink (even lower saturation)
    "#d0d3d3",  # Muted Teal (even lower saturation)
    "#dfd9d9",  # Muted Green (even lower saturation)
    "#d3d6d6",  # Muted Peach (even lower saturation)
    "#cacaca",  # Muted Silver (even lower saturation)
]



from setuptools import setup, find_packages

from pathlib import Path
long_description = Path("../readme.md").read_text()

setup(
    name="sec_parsers",
    author="John Friedman",
    version="0.513",
    description = "A package to parse SEC filings",
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=[
        'lxml', 
    ],
)


import re
from sec_parsers.xml_helper import get_all_text

#TODO happy with this file so far

# simple for now
def is_paragraph(text):
    periods = text.count('.')
    commas = text.count(',')
    if periods > 0:
        return True
    
    return False

# TODO: improve this
def detect_bullet_point(string):
    """e.g. •"""
    if any(char in string.strip() for char in ['•','●','●','●','●','•','·','◦']):
        return True
    return False

def detect_signatures(string):
    """e.g. Signatures"""
    match = re.search(r"^SIGNATURES$",string, re.IGNORECASE)
    if match:
        return True
    
    return False

def detect_style_from_string(string):
    def detect_emphasis_capitalization(string):
        """Seen in amazon's 2024 10k e.g. We Have Foreign Exchange Risk"""

        # TODO: FIX
        if string.lower() in ['None','None.','Omitted.', 'Not Applicable.']:
            return False
        
        words = string.split()
        if not words:
            return False
        for word in words:
            if word.lower() in ["of",'to','a','and','by','in','the','or','on','for','as','with','that','but','not','so','yet','an','at','off','per','up','via']:
                continue
            # check for numbers, e.g. 2021, but not ITEM2021. ITEM2021 counts as upper
            elif not any(char.isalpha() for char in word):
                continue

            if not word[0].isupper():
                return False
        return True
    
    def detect_item(string):
        """e.g. Item 1A. Risk Factors"""
        match = re.search(r"^Item\s+\d+[A-Z]{0,}",string, re.IGNORECASE)
        if match:
            return True
        return False
    
    def detect_part(string):
        """e.g. Part I"""
        match = re.search(r"^Part\s+\w+",string, re.IGNORECASE)
        if match:
            return True
        return False
        
    # deprecated for now
    def level_detection(string):
        """e.g. amazon Level 1"""
        match = re.search(r"^Level\s+\d+",string)
        if match:
            return True
        return False
        

    def note_detection(string):
        """e.g. Note 1"""
        match = re.search(r"^Note\s+\d+",string, re.IGNORECASE)
        if match:
            return True
        return False
    
    def all_caps(string):
        """e.g. FORM 10-K SUMMARY"""
        if string.isupper():
            return True
        # may cause issue
        elif string.isdigit():
            return True
        return False
    
    def detect_page_number(string):
        if re.search(r"^\d+$",string, re.IGNORECASE):
            return True
        elif re.search(r'F-\d+$',string, re.IGNORECASE):
            return True
        else:
            return False
    
    # PREPROCESSING WIP
    string = string.strip()
    # ORDER MATTERS
    if detect_part(string):
        return 'part;'
    elif detect_item(string):
        return 'item;'
    elif detect_signatures(string):
        return 'signatures;'
    elif detect_page_number(string):
        return 'page number;'
    elif detect_bullet_point(string):
        return 'bullet point;'
    elif all_caps(string):
        return 'all caps;'
    elif detect_emphasis_capitalization(string):
        return 'emphasis;'
    elif note_detection(string):
        return 'note;'
    else:
        return ''
    

# New
def detect_hidden_element(element):
    if element.get('style'):
        if 'display:none' in re.sub(' ','',element.get('style')):
            return True
    return False

    
# add multiple so italic and bold
def detect_style_from_element(element):
    def detect_bold_from_css(element):
        """Detects bold from css"""
        if element.get('style'):
            if 'font-weight:bold' in element.get('style'):
                return 'font-weight:bold;'
            # change to be any font weight greater than 400
            elif 'font-weight:700' in element.get('style'):
                return 'font-weight:700;'
        return ''
    
    def detect_from_html(element):
        element_style = ''
        ancestors = list(element.iterancestors())
        # include self
        ancestors = ancestors + [element]
        # detect bold from ancestors
        for ancestor in ancestors:
            tag_list = ['b','strong','em','i','u']
            if ancestor.tag in tag_list:
                element_style += f'{ancestor.tag}-tag;'
            
        return element_style
    
    def detect_underline_from_css(element):
        """Detects underline from css"""
        if element.get('style'):
            if 'text-decoration:underline' in element.get('style'):
                return 'text-decoration:underline;'
        return ''
    
    def detect_italic_from_css(element):
        """Detects italic from css"""
        if element.get('style'):
            if 'font-style:italic' in element.get('style'):
                return 'font-style:italic;'
        return ''

    
    # check it or descendants have text
    text = get_all_text(element).strip()
    if len(text) == 0:
        return ''

    
    style = ''
    style += detect_bold_from_css(element)
    style += detect_from_html(element)
    style += detect_underline_from_css(element)
    style += detect_italic_from_css(element)
    return style
    

# needs work 
def detect_table(table):
    """Detects if table or header disguised as a table"""
    if table.tag != 'table':
        return False
    
    tr_list = table.xpath('//tr')
    if len(tr_list) > 3:
        return True
    text = get_all_text(table)

    if detect_bullet_point(text):
        return False
    number_count = len(re.findall(r'\d', text))
    if number_count > 5:
        return True
    
    char_count = len(text)
    if char_count > 400:
        return True
    
    return False


def detect_table_of_contents(element):
    toc_type = 'not-toc'
    """Detects if a table is likely to be a table of contents."""

    # toc - needs seperate parser if no links
    num_items = len(re.findall('Item(\s+|$|\n)', get_all_text(element), re.IGNORECASE))
    if num_items > 9:
        toc_type = 'toc'

    # linked toc
    # links = element.find_all('a')
    # if len(links) > 5:
    #     toc_type = 'toc-links'

    return toc_type

def detect_toc_link(node):
    """Detects if a node is a table of contents link."""
    if node.tag == 'a':
        text = get_all_text(node)
        if text.lower() in ['table of contents','toc']:
            return True
    return False

def detect_link(node):
    """Detects if a node is a link."""
    if node.tag == 'a':
        return True
    return False

def detect_image(node):
    """Detects if a node is an image."""
    if node.tag == 'img':
        return True
    return False


import webbrowser
import tempfile
from lxml import html, etree
import re

# TODO: lookover and refactor

 # Used for conversion to xml #

# this probably slows down code a lot. if runtime affected, look here. 
def get_elements_between_elements(root, start_element, end_element):
    """WIP"""
    elements = []
    in_between_bool = False
    for element in root.iter():
        if element == start_element:
            in_between_bool = True
        elif element == end_element:
            in_between_bool = False
        elif in_between_bool:
            elements.append(element)

    return elements

def get_text_between_elements(root, start_element, end_element):
    """TODO: format text nicer"""
    elements = get_elements_between_elements(root, start_element, end_element)
    text = ""
    for element in elements:
        element_text = get_text(element).strip()
        if element_text != '':
            text += element_text + '\n'

    return text

# used for parsed html #

def element_has_text(element):
    text = element.text
    if text:
        text = text.strip()
        if text != '':
            return True
    return False

def element_has_tail(element):
    tail = element.tail
    if tail:
        tail = tail.strip()
        if tail != '':
            return True
    return False

def check_if_is_first_child(element):
    """Get first child of element"""
    previous_element = element.getprevious()

    if previous_element is None:
        parent = element.getparent()
        if not element_has_text(parent):
            return True
        else:
            return False
    elif len(previous_element.getchildren()) > 0:
        return False

    return False
    
def get_text(element):
    """Get text from element including tail"""
    text = ''
    if element.text:
        text += element.text

    if element.tail:
        text += ' ' + element.tail
    return text

# check includes tail
def get_all_text(node):
    """Get all text from element including children. Make include tail"""
    text = ''.join(node.itertext())

    return text


# visualization
def remove_style(element):
    """Removes the style attribute from an element."""
    element.attrib.pop('style', None)

def remove_background_color(element):
    """Removes the background color from an element."""
    current_style = element.get('style')
    if current_style:
        # remove background color from style if exists
        current_style = re.sub(r'background-color:.*?($|;)','',current_style)
        element.set('style', current_style)
    else:
        element.set('style', '')

def set_background_color(element, color):
    """Sets the background color for an element."""
    current_style = element.get('style')
    if current_style:
        new_style = f'{current_style}; background-color: {color}'
        element.set('style', new_style)
    else:
        element.set('style', f'background-color: {color}')
    
def remove_background(element):
    return

# Visualization #

def open_tree(tree):
    """Opens a lxml tree in a web browser."""
    with tempfile.NamedTemporaryFile('w', delete=False, suffix='.html', encoding="utf-8-sig") as f:
        data = html.tostring(tree).decode("utf-8-sig")
        f.write(data)

    url = 'file://' + f.name
    webbrowser.open(url)



from collections import defaultdict

# TODO: Add more comments to the functions, refactor

# Module to take parsed headers and determine the hierarchy of the headers

# Highly WIP
def get_hierarchy_from_list(parsing_list):
    seen = set()
    result = []
    for parsing_string in parsing_list:
        if parsing_string not in seen:
            seen.add(parsing_string)
            result.append(parsing_string)
    return result

def split_list(parsing_list):
    lsts = []
    lst = []
    for parsing_string in parsing_list:
        if ((parsing_string == 'item;') or (parsing_string == 'part;')):
            if len(lst) > 0:
                lsts.append(lst)
            lst = []
        else:
            lst.append(parsing_string)

    return lsts

def get_hierarchy_from_lists(lsts):
    # Initialize the counter dictionary
    pair_counts = defaultdict(int)

    # Update counts for each pair
    for lst in lsts:
        for i in range(len(lst) - 1):
            pair = (lst[i], lst[i + 1])
            pair_counts[pair] += 1

    # Convert pair counts to a list of tuples and sort by count
    sorted_pairs = sorted(pair_counts.items(), key=lambda x: x[1], reverse=True)

    # Initialize a set to keep track of visited elements
    visited = set()
    hierarchy = []

    # Build the hierarchy based on sorted pairs
    for (a, b), count in sorted_pairs:
        if a not in visited:
            hierarchy.append(a)
            visited.add(a)
        if b not in visited:
            hierarchy.append(b)
            visited.add(b)

    return hierarchy

# WIP
def get_hierarchy(elements):
    """Calculates the hierarchy."""
    # get lists item --> item, item --> part
    lsts = split_list(elements)
    # get hiereachy lists
    hierarchy_lists = [get_hierarchy_from_list(lst) for lst in lsts]
    # use lists to get hierarchy
    hierarchy = get_hierarchy_from_lists(hierarchy_lists)
    # insert item at beginning
    hierarchy.insert(0,'item;')
    # insert part before item
    hierarchy.insert(0,'part;')
    return hierarchy


def get_preceding_elements(element_list, element):
    """Get the elements that precede the given element in the list."""
    if element in element_list:
        element_index = element_list.index(element)
        return element_list[:element_index]
    else:
        return []

def find_last_index(lst1, lst2):
  """Find the last index of lst1 in lst2."""
  last_index = -1
  for i in range(len(lst1)):
    if lst1[i] in lst2:
      last_index = i
  return last_index

